from rest_framework import serializers

from .models import Currency, CurrencyNotFoundError


class CurrencySerializer(serializers.Serializer):
    code = serializers.CharField(read_only=True)
    name = serializers.CharField(read_only=True)
    currency_name = serializers.CharField(read_only=True)
    exponent = serializers.IntegerField(read_only=True)
    number = serializers.IntegerField(read_only=True)
    value = serializers.CharField(read_only=True)
    symbol = serializers.SerializerMethodField()

    def get_symbol(self, obj):
        return obj.symbol

    def create(self, validated_data):
        return Currency(code=validated_data.get('code'))




from django.utils.translation import ugettext_lazy as _, pgettext_lazy

DIMENSION_FAMILIES = {
    'dimensionless': _('dimensionless'),
    'meter': _('length'),
    'meter ** 2': _('surface'),
    'meter ** 3': _('volume'),
    'second': _('time'),
    'gram': _('mass'),
    'meter / second': _('speed'),
    'meter / second ** 2': _('acceleration'),
    'meter / gram': _('meter per gram'),
    'meter ** 2 / second ** 2': _('absorption'),
    'gram * meter / second ** 2': _('force'),
    'gram * meter ** 2 / second ** 2': _('energy'),
    'ampere': _('electric current'),
    'ampere * second': _('electric charge'),
    'ampere * second / mole': _('elementary charge'),
    'gram * meter ** 2 / ampere ** 2 / second ** 2': _('electromagnetic inductance'),
    'gram * meter ** 2 / ampere ** 2 / second ** 3': _('electromagnetic resistance'),
    'ampere ** 2 * second ** 3 / gram / meter ** 2': _('electrical conductance'),
    'gram * meter ** 2 / ampere / second ** 3': _('electric potential difference'),
    'gram * meter / ampere / second ** 3': _('electric field'),
    'gram / meter': _('linear mass density'),
    'gram / second ** 3': _('intensity'),
    'kelvin': _('temperature'),
    'mole': _('amount of substance'),
    '1 / mole': _('density of particle'),
    'mole / second': _('catalytic activity'),
    'bit': _('quantity of information'),
    'bit / pixel': _('density of information'),
    'ampere * meter ** 2': _('magnetic moment'),
    'ampere * meter * second': _('electric moment'),
    'gram * meter ** 2 / kelvin / second ** 2': _('energy'),
    'ampere * meter ** 2 * second': _('magnetic moment'),
    'candela': _('luminous intensity'),
    'ampere ** 2 * second ** 4 / gram / meter ** 2': _('electrical capacitance'),
    'gram * meter ** 3 / ampere ** 2 / second ** 4': _('electric force'),
    'count / second': _('frequency'),
    'gram * meter ** 2 / second': _('electromagnetic action'),
    'gram * meter ** 4 / second ** 3': _('power per unit per wavelength interval'),
    'gram ** 0.5 * meter ** 1.5 / second': _('electrostatic unit of charge'),
    'gram ** 0.5 / meter ** 0.5 / second': _('magnetic induction'),
    'radian': _('angle'),
    'candela / meter ** 2': _('luminance'),
    'gram / second ** 2': _('radiation'),
    'candela * radian ** 2': _('luminous flux'),
    'candela * radian ** 2 / meter ** 2': _('illuminance'),
    'gram * meter ** 2 / ampere / second ** 2': _('magnetic flux quantum'),
    'mole / meter ** 3': _('concentration'),
    'gram * meter ** 2 / kelvin / mole / second ** 2': _('energy per temperature increment per mole'),
    'neper': _('neper'),
    'meter ** 3 / gram / second ** 2': _('newtonian constant of gravitation'),
    'gram / meter / second': _('viscosity'),
    'gram / meter / second ** 2': _('pressure'),
    '1 / meter': _('reciprocal length'),
    'radian / second': _('angular speed'),
    'meter * second / gram': _('reciprocal viscosity'),
    'ampere * second / gram': _('radiation exposure'),
    'radian ** 2': _('solid degree'),
    'gram * meter ** 2 / second ** 3': _('volumetric flow rate'),
    'gram / ampere / second ** 2': _('magnetic field'),
    'gram / meter ** 3': _('mass density')
}

UNIT_EXTENDED_DEFINITION = {
    'K_alpha_Cu_d_220': {
        'name': _('Copper Kα'),
        'symbol': 'Cu Kα',
        'family': _('dimensionless')
    },
    'K_alpha_Mo_d_220': {
        'name': _('Molybdenum K-α'),
        'symbol': 'MO Kα',
        'family': _('dimensionless')
    },
    'K_alpha_W_d_220': {
        'name': _('Tungsten K-α'),
        'symbol': _('W Kα', ),
        'family': _('dimensionless')
    },
    'RKM': {
        'name': _('RKM'),
        'symbol': 'RKM',
        'obsolete': True,
        'family': _('absorption')
    },
    'UK_force_ton': {
        'name': _('long ton-force'),
        'symbol': 'long ton-force',
        'obsolete': True,
        'family': _('force')

    },
    'UK_hundredweight': {
        'name': 'long hundredweight',
        'symbol': 'long hundredweight',
        'obsolete': True,
        'family': _('mass')
    },
    'UK_ton': {
        'name': _('long ton'),
        'symbol': 'long ton',
        'obsolete': True,
        'family': _('mass')
    },
    'US_force_ton': {
        'name': _('short ton force'),
        'symbol': 'short ton-force',
        'obsolete': True,
        'family': _('force')
    },
    'US_hundredweight': {
        'name': _('short hundredweight'),
        'symbol': 'short hundredweight',
        'obsolete': True,
        'family': _('mass')
    },
    'US_ton': {
        'name': _('short ton'),
        'symbol': 'short ton',
        'obsolete': True,
        'family': _('mass')
    },
    'US_international_ampere': {
        'name': _('US International Ampere'),
        'symbol': 'US international A',
        'obsolete': True,
        'family': _('electric current')
    },
    'US_international_ohm': {
        'name': _('US international Ohm'),
        'symbol': 'US international Ω',
        'obsolete': True,
        'family': _('electric charge')
    },
    'US_international_volt': {
        'name': _('US international Volt'),
        'symbol': 'US international V',
        'obsolete': True,
        'family': _('electric potential difference')
    },
    'US_therm': {
        'name': _('US therm'),
        'symbol': 'thm',
        'family': _('heat energy')
    },
    'abampere': {
        'name': _('abampere'),
        'symbol': 'abA',
        'family': _('electric current')
    },
    'abcoulomb': {
        'name': _('abcoulomb'),
        'symbol': 'abC',
        'family': _('electromagnetic inductance')
    },
    'aberdeen': {
        'name': _('Aberdeen'),
        'symbol': 'aberdeen',
        'family': _('linear mass density')
    },
    'abfarad': {
        'name': _('abfarad'),
        'symbol': 'abF',
        'family': _('electrical capacitance')
    },
    'abhenry': {
        'name': _('abhenry'),
        'symbol': 'henry',
        'family': _('electrical inductance'),
    },
    'abohm': {
        'name': _('abohm'),
        'symbol': 'abohm',
        'family': _('electric charge')
    },
    'absiemens': {
        'name': _('absiemens'),
        'symbol': 'absiemens',
        'family': _('electrical conductance')
    },
    'abvolt': {
        'name': _('abvolt'),
        'symbol': 'abV',
        'family': _('electric potential difference')
    },
    'acre': {
        'name': _('acre'),
        'symbol': 'acre',
        'family': _('surface')
    },
    'acre_foot': {
        'name': _('acre-foot'),
        'symbol': 'acre-foot',
        'family': _('volume')
    },
    'ampere': {
        'name': _('ampere'),
        'symbol': 'A',
        'family': _('electric current'),
    },
    'ampere_turn': {
        'name': _('ampere-turn'),
        'symbol': 'At',
        'family': _('electric current'),
    },
    'angstrom': {
        'name': _('angstrom'),
        'symbol': 'Å',
        'family': _('length')
    },
    'angstrom_star': {
        'name': _('angstrom star'),
        'symbol': 'Å*',
        'family': _('length')
    },
    'apothecary_dram': {
        'name': _('apothecary dram'),
        'symbol': 'dr',
        'family': _('mass')
    },
    'apothecary_ounce': {
        'name': _('apothecary ounce'),
        'symbol': 'oz',
        'family': _('mass')
    },
    'apothecary_pound': {
        'name': _('apothecary pound'),
        'symbol': 'apothecary pound',
        'family': _('mass')
    }, 'arcminute': {
        'name': _('minute of arc'),
        'symbol': 'arcmin',
        'family': _('dimensionless')
    }, 'arcsecond': {
        'name': _('second of arc'),
        'symbol': 'arcsec',
        'family': _('dimensionless')
    }, 'are': {
        'name': _('are'),
        'symbol': 'are',
        'family': _('surface')
    }, 'astronomical_unit': {
        'name': _('astronomical unit'),
        'symbol': 'AU',
        'family': _('length')
    }, 'atmosphere_liter': {
        'name': _('atmosphere liter'),
        'symbol': 'atmosphere liter',
        'family': _('volume')
    }, 'atomic_mass_constant': {
        'name': _('atomic mass constant'),
        'symbol': 'Da',
        'family': _('constant')
    },
    'atomic_unit_of_current': {
        'name': _('atomic unit of current'),
        'symbol': '',
        'family': _('ampere')
    },
    'atomic_unit_of_electric_field': {
        'name': _('atomic unit of electric field'),
        'symbol': '',
        'family': _('electric field')
    }, 'atomic_unit_of_force': {
        'name': _('atomic unit of force'),
        'symbol': 'atomic unit of force',
        'family': _('force')
    }, 'atomic_unit_of_intensity': {
        'name': _('atomic unit of intensity'),
        'symbol': 'atomic unit of intensity',
        'family': _('intensity')
    }, 'atomic_unit_of_temperature': {
        'name': _('atomic unit of temperature'),
        'symbol': 'atomic unit of temperature',
        'family': _('temperature')
    }, 'atomic_unit_of_time': {
        'name': _('atomic unit of time'),
        'symbol': 'atomic unit of time',
        'family': _('time')
    }, 'avogadro_constant': {
        'name': _('Avogadro constant'),
        'symbol': 'L',
        'family': _('constant')
    }, 'avogadro_number': {
        'name': _('Avogadro number'),
        'symbol': '',
        'family': _('dimensionless'),
    }, 'bag': {
        'name': _('bag'),
        'symbol': 'bag',
        'family': _('quantity')
    }, 'bar': {
        'name': _('bar'),
        'symbol': 'bar',
        'family': _('pressure')
    }, 'barn': {
        'name': _('barn'),
        'symbol': 'b',
        'family': _('surface')
    }, 'barrel': {
        'name': _('barrel'),
        'symbol': 'barrel',
        'family': 'volume'
    }, 'barye': {
        'name': _('barrye'),
        'symbol': 'Ba',
        'family': _('pressure')
    },
    'baud': {
        'name': _('baud'),
        'symbol': 'Bd',
        'family': _('speed')
    },
    'becquerel': {
        'name': _('becquerel'),
        'symbol': 'Bq',
        'family': _('radioactivity')
    },
    'beer_barrel': {
        'name': _('beer barrel'),
        'symbol': 'beer barrel',
        'family': _('volume')
    },
    'bel': {
        'name': _('bel'),
        'symbol': 'B',
        'family': _('field quantity')
    },
    'biot': {
        'name': _('biot'),
        'symbol': 'biot',
        'family': _('dimensionless')
    },
    'biot_turn': {
        'name': _('biot turn'),
        'symbol': 'biot-turn',
        'family': _('dimensionless')
    },
    'bit': {
        'name': _('bit'),
        'symbol': 'bit',
        'family': _('quantity of information')
    },
    'bits_per_pixel': {
        'name': _('bits per pixel'),
        'symbol': 'bpp',
        'family': _('density of information')
    },
    'board_foot': {
        'name': _('board foot'),
        'symbol': 'board-foot',
        'family': _('volume')
    },
    'bohr': {
        'name': _('bohr'),
        'symbol': 'bohr',
        'family': _('length')
    },
    'bohr_magneton': {
        'name': _('bohr magneton'),
        'symbol': 'μB',
        'family': _('magnetic moment')
    },
    'boiler_horsepower': {
        'name': _('boiler horsepower'),
        'symbol': 'boiler hp',
        'family': _('power')
    },
    'boltzmann_constant': {
        'name': _('boltzmann constant'),
        'symbol': 'kB',
        'family': _('constant')
    },
    'british_thermal_unit': {
        'name': _('british thermal unit'),
        'symbol': 'Btu',
        'family': _('energy')
    },
    'buckingham': {
        'name': _('buckingham'),
        'symbol': 'B',
        'family': _('magnetic moment')
    },
    'bushel': {
        'name': _('bushel'),
        'symbol': 'bsh',
        'family': _('volume')
    },
    'byte': {
        'name': _('byte'),
        'symbol': 'byte',
        'family': _('quantity of information')
    },
    'cables_length': {
        'name': _('cable length'),
        'symbol': 'cable',
        'family': _('length')
    },
    'calorie': {
        'name': _('calorie'),
        'symbol': 'calorie',
        'family': _('energy')
    },
    'candela': {
        'name': _('candela'),
        'symbol': 'cd',
        'family': _('luminous intensity')
    },
    'carat': {
        'name': _('carat'),
        'symbol': 'ct',
        'family': _('mass')
    },
    'centimeter_H2O': {
        'name': _('centimeter H20'),
        'symbol': 'cm H2O',
        'family': _('pressure')
    },
    'centimeter_Hg': {
        'name': _('centimeter Hg'),
        'symbol': 'cm Hg',
        'family': _('pressure')
    },
    'century': {
        'name': _('century'),
        'symbol': 'c.',
        'family': _('time')
    },
    'chain': {
        'name': _('chain'),
        'symbol': 'chain',
        'family': _('length')
    },
    'cicero': {
        'name': _('cicero'),
        'symbol': 'cicero',
        'family': _('length')
    },
    'circular_mil': {
        'name': _('circular mil'),
        'symbol': 'mils',
        'family': _('surface')
    },
    'classical_electron_radius': {
        'name': _('classical electron radius'),
        'symbol': 're',
        'family': _('length')
    },
    'clausius': {
        'name': _('clausius'),
        'symbol': 'clausius',
        'family': _('energy')
    },
    'common_year': {
        'name': _('common year'),
        'symbol': '',
        'family': _('time')
    },
    'conductance_quantum': {
        'name': _('conductance quantum'),
        'symbol': 'G0',
        'family': _('electrical conductance')
    },
    'conventional_ampere_90': {
        'name': _('conventional ampere'),
        'symbol': 'A90',
        'family': _('electric current')
    },
    'conventional_coulomb_90': {
        'name': _('conventional coulomb'),
        'symbol': 'C90',
        'family': _('electromagnetic inductance')
    },
    'conventional_farad_90': {
        'name': _('conventional farad'),
        'symbol': 'F90',
        'family': _('eletrical capacitance')
    },
    'conventional_henry_90': {
        'name': _('conventional henry'),
        'symbol': 'H90',
        'family': _('electrical inductance'),
    },
    'conventional_josephson_constant': {
        'name': _('conventional josephson constant'),
        'symbol': 'KJ',
        'family': _('constant'),
    },
    'conventional_ohm_90': {
        'name': _('conventional ohm'),
        'symbol': 'Ω90',
        'family': _('electric charge')
    },
    'conventional_volt_90': {
        'name': _('conventional volt'),
        'symbol': 'V90',
        'family': _('electric potential difference')
    },
    'conventional_von_klitzing_constant': {
        'name': _('conventional Von Klitzing constant'),
        'symbol': 'RK',
        'family': _('constant'),
    },
    'conventional_watt_90': {
        'name': _('conventional Watt'),
        'symbol': 'W90',
        'family': _('power'),
    },
    'coulomb': {
        'name': _('coulomb'),
        'symbol': 'C',
        'family': _('electromagnetic inductance')
    }, 'coulomb_constant': {
        'name': _('coulomb constant'),
        'symbol': 'K',
        'family': _('constant')
    },
    'count': {
        'name': _('count'),
        'symbol': '',
        'family': _('dimensionless')
    },
    'counts_per_second': {
        'name': _('counts per second'),
        'symbol': 'counts/s',
        'family': _('frequency')
    },
    'css_pixel': {
        'name': _('CSS pixel'),
        'symbol': 'px',
        'family': _('length')
    },
    'cubic_centimeter': {
        'name': _('cubic centimeter'),
        'symbol': 'cm³',
        'family': _('volume')
    },
    'cubic_foot': {
        'name': _('cubic foot'),
        'symbol': 'cubic foot',
        'family': _('volume')
    },
    'cubic_inch': {
        'name': _('cubic inch'),
        'symbol': 'cubic inch',
        'family': _('volume')
    }, 'cubic_yard': {
        'name': _('cubic yard'),
        'symbol': 'cubic yard',
        'family': _('volume')
    }, 'cup': {
        'name': _('cup'),
        'symbol': 'cup',
        'family': _('volume')
    }, 'curie': {
        'name': _('curie'),
        'symbol': 'Ci',
        'family': _('frequency')
    }, 'dalton': {
        'name': _('dalton'),
        'symbol': 'Da',
        'family': _('mass')
    }, 'darcy': {
        'name': _('darcy'),
        'symbol': 'D',
        'family': _('permeability')
    }, 'day': {
        'name': _('day'),
        'symbol': pgettext_lazy('unit', 'day'),
        'family': _('time')
    }, 'debye': {
        'name': _('debye'),
        'symbol': 'D',
        'family': _('electric moment')
    }, 'decade': {
        'name': _('decade'),
        'symbol': pgettext_lazy('symbol', 'decade'),
        'family': _('time')
    }, 'decibel': {
        'name': _('decibel'),
        'symbol': 'dB',
        'family': _('field quantity')
    },'degree': {
        'name': _('degree'),
        'symbol': '°',
        'family': _('dimensionless')
    }, 'degree_Celsius': {
        'name': _('degree Celsius'),
        'symbol': '°C',
        'family': _('temperature')
    }, 'degree_Fahrenheit': {
        'name': _('degree Fahrenheit'),
        'symbol': '°F',
        'family': _('temperature')
    }, 'degree_Rankine': {
        'name': _('degree Rankine'),
        'symbol': '°R',
        'family': _('temperature')
    }, 'degree_Reaumur': {
        'name': _('degree Reaumur'),
        'symbol': '°Re',
        'family': _('temperature')
    }, 'denier': {
        'name': _('denier'),
        'symbol': 'D',
        'family': _('linear mass density')
    }, 'didot': {
        'name': _('didot'),
        'symbol': 'didot',
        'family': _('length')
    }, 'dirac_constant': {
        'name': _('Dirac constant'),
        'symbol': 'ħ',
        'family': _('constant')
    }, 'dram': {
        'name': _('dram'),
        'symbol': 'dr',
        'family': _('mass')
    }, 'dry_barrel': {
        'name': _('dry barrel'),
        'symbol': 'dry barrel',
        'family': _('volume')
    }, 'dry_gallon': {
        'name': _('dry gallon'),
        'symbol': 'dry gallon',
        'family': _('volume')
    }, 'dry_pint': {
        'name': _('dry pint'),
        'symbol': 'dry pint',
        'family': _('volume')
    }, 'dry_quart': {
        'name': _('dry quart'),
        'symbol': 'dry quart',
        'family': _('volume')
    }, 'dtex': {
        'name': _('decitex'),
        'symbol': 'dtex',
        'family': _('linear mass density')
    }, 'dyne': {
        'name': _('dyne'),
        'symbol': 'dyn',
        'family': _('force')
    }, 'electrical_horsepower': {
        'name': _('eletcrical horsepower'),
        'symbol': 'HP',
        'family': _('power')
    }, 'electron_g_factor': {
        'name': _('electron G-factor'),
        'symbol': 'electron G-factor',
        'family': _('dimensionless')
    }, 'electron_mass': {
        'name': _('electron mass'),
        'symbol': 'me',
        'family': _('mass')
    }, 'electron_volt': {
        'name': _('electronvolt'),
        'symbol': 'eV',
        'family': _('electric potential difference')
    }, 'elementary_charge': {
        'name': _('elementary charge'),
        'symbol': 'e',
        'family': _('electromagnetic inductance')
    }, 'entropy_unit': {
        'name': _('entropy unit'),
        'symbol': 'S',
        'family': _('energy by temperature')
    }, 'enzyme_unit': {
        'name': _('enzyme unit'),
        'symbol': 'U',
        'family': _('catalytic activity')
    }, 'eon': {
        'name': _('eon'),
        'symbol': 'eon',
        'family': _('time')
    }, 'erg': {
        'name': _('erg'),
        'symbol': 'erg',
        'family': _('energy')
    }, 'farad': {
        'name': _('farad'),
        'symbol': 'F',
        'family': _('electrical capacitance')
    }, 'faraday': {
        'name': _('faraday'),
        'symbol': 'faraday',
        'family': _('electric charge'),
    }, 'faraday_constant': {
        'name': _('faraday constant'),
        'symbol': 'F',
        'family': _('constant'),
    }, 'fathom': {
        'name': _('fathom'),
        'symbol': 'fathom',
        'family': _('length'),
    }, 'fermi': {
        'name': _('fermi'),
        'symbol': 'fm',
        'family': _('length'),
    }, 'fifteen_degree_calorie': {
        'name': _('fifteen degree calorie'),
        'symbol': 'fifteen degree calorie',
        'family': _('energy'),
    }, 'fifth': {
        'name': _('fifth'),
        'symbol': 'fifth',
        'family': _('volume'),
    }, 'fine_structure_constant': {
        'name': _('fine structure constant'),
        'symbol': 'α',
        'family': _('constant'),
    }, 'first_radiation_constant': {
        'name': _('first radiation constant'),
        'symbol': 'c1',
        'family': _('constant'),
    }, 'fluid_dram': {
        'name': _('fluid dram'),
        'symbol': 'fluid dr',
        'family': _('mass')
    }, 'fluid_ounce': {
        'name': _('fluid ounce'),
        'symbol': 'fluid ounce',
        'family': _('mass')
    }, 'foot': {
        'name': _('foot'),
        'symbol': 'foot',
        'family': _('length')
    }, 'foot_H2O': {
        'name': _('foot H2O'),
        'symbol': 'foot H2O',
        'family': _('pressure')
    }, 'foot_per_second': {
        'name': _('foot per second'),
        'symbol': 'foot per second',
        'family': _('speed')
    }, 'foot_pound': {
        'name': _('foot pound'),
        'symbol': 'foot pound',
        'family': _('energy')
    }, 'force_gram': {
        'name': _('gram-force'),
        'symbol': 'gram-force',
        'obsolete': True,
        'family': _('force')
    }, 'force_kilogram': {
        'name': _('kilogram-force'),
        'symbol': 'kg-force',
        'obsolete': True,
        'family': _('force')
    }, 'force_long_ton': {
        'name': _('long ton-force'),
        'symbol': 'long ton-force',
        'obsolete': True,
        'family': _('force')
    }, 'force_metric_ton': {
        'name': _('tonne-force'),
        'symbol': 'tf',
        'obsolete': True,
        'family': _('force')
    }, 'force_ounce': {
        'name': _('ounce-force'),
        'symbol': 'ounce force',
        'obsolete': True,
        'family': _('force')
    }, 'force_pound': {
        'name': _('pound-force'),
        'symbol': 'pound force',
        'obsolete': True,
        'family': _('force')
    }, 'force_ton': {
        'name': _('tonne-force'),
        'symbol': 'tf',
        'obsolete': True,
        'family': _('force')
    }, 'fortnight': {
        'name': _('fortnight'),
        'symbol': 'fortnight',
        'family': _('time')
    }, 'franklin': {
        'name': _('franklin'),
        'symbol': 'Fr',
        'family': _('electrostatic unit of charge')
    }, 'furlong': {
        'name': _('furlong'),
        'symbol': 'furlong',
        'family': _('length')
    }, 'galileo': {
        'name': _('galileo'),
        'symbol': 'Gal',
        'family': _('acceleration')
    }, 'gallon': {
        'name': _('gallon'),
        'symbol': 'gal',
        'family': _('volume')
    }, 'gamma': {
        'name': _('gamma'),
        'symbol': 'γ',
        'family': _('constant')
    }, 'gamma_mass': {
        'name': _('gamma'),
        'symbol': 'γ',
        'family': _('mass')
    }, 'gauss': {
        'name': _('gauss'),
        'symbol': 'G',
        'family': _('magnetic induction')
    }, 'gilbert': {
        'name': _('gilbert'),
        'symbol': 'Gb',
        'family': _('electric current'),
    }, 'gill': {
        'name': _('gill'),
        'symbol': 'gill',
        'family': _('volume'),
    }, 'grade': {
        'name': _('grade'),
        'symbol': 'gr',
        'family': _('angle'),
    }, 'grain': {
        'name': _('grain'),
        'symbol': 'grain',
        'family': _('mass'),
    }, 'gram': {
        'name': _('gram'),
        'symbol': 'g',
        'family': _('mass'),
    }, 'gray': {
        'name': _('gray'),
        'symbol': 'Gy',
        'family': _('absorption'),
    }, 'gregorian_year': {
        'name': _('gregorian year'),
        'symbol': pgettext_lazy('unit', 'gregorian year'),
        'family': _('time')
    }, 'hand': {
        'name': _('hand'),
        'symbol': 'hand',
        'family': _('length'),
    }, 'hartree': {
        'name': _('hartree'),
        'symbol': 'Eh',
        'family': _('energy'),
    }, 'hectare': {
        'name': _('hectare'),
        'symbol': 'ha',
        'family': _('surface'),
    }, 'henry': {
        'name': _('henry'),
        'symbol': 'henry',
        'family': _('electrical inductance'),
    }, 'hertz': {
        'name': _('hertz'),
        'symbol': 'Hz',
        'family': _('frequency'),
    }, 'hogshead': {
        'name': _('hogshead'),
        'symbol': 'hhd',
        'family': _('volume'),
    }, 'horsepower': {
        'name': _('horsepower'),
        'symbol': 'HP',
        'family': _('power'),
    }, 'hour': {
        'name': _('hour'),
        'symbol': 'h',
        'family': _('time'),
    }, 'hundredweight': {
        'name': _('hundredweight'),
        'symbol': 'cwt',
        'family': _('mass'),
    }, 'impedance_of_free_space': {
        'name': _('impedance of free space'),
        'symbol': 'Z0',
        'family': _('electromagnetic resistance'),
    }, 'imperial_barrel': {
        'name': _('imperial barrel'),
        'symbol': 'imperial barrel',
        'family': _('volume'),
    }, 'imperial_bushel': {
        'name': _('imperial bushel'),
        'symbol': 'bsh',
        'family': _('volume'),
    }, 'imperial_cup': {
        'name': _('imperial cup'),
        'symbol': 'imperial cup',
        'family': _('volume'),
    }, 'imperial_fluid_drachm': {
        'name': _('imperial fluid drachm'),
        'symbol': 'ʒ',
        'family': _('mass'),
    }, 'imperial_fluid_ounce': {
        'name': _('imperial fluid ounce'),
        'symbol': 'fl oz',
        'family': _('volume'),
    }, 'imperial_fluid_scruple': {
        'name': _('imperial fluid scruple'),
        'symbol': '℈',
        'family': _('volume'),
    }, 'imperial_gallon': {
        'name': _('imperial gallon'),
        'symbol': 'imperial gallon',
        'family': _('volume'),
    }, 'imperial_gill': {
        'name': _('imperial gill'),
        'symbol': 'imperial gill',
        'family': _('volume'),
    }, 'imperial_minim': {
        'name': _('imperial minim'),
        'symbol': 'min',
        'family': _('volume'),
    }, 'imperial_peck': {
        'name': _('imperial peck'),
        'symbol': 'peck',
        'family': _('volume')
    }, 'imperial_pint': {
        'name': _('imperial pint'),
        'symbol': 'imperial pint',
        'family': _('volume')
    }, 'imperial_quart': {
        'name': _('imperial quart'),
        'symbol': 'imperial quart',
        'family': _('volume')
    }, 'inch': {
        'name': _('inch'),
        'symbol': 'inch',
        'family': _('length')
    }, 'inch_H2O_39F': {
        'name': _('inch H20 39F'),
        'symbol': 'inch H20 39F',
        'family': _('pressure')
    }, 'inch_H2O_60F': {
        'name': _('inch H20 60F'),
        'symbol': 'inch H20 60F',
        'family': _('pressure')
    }, 'inch_Hg': {
        'name': _('inch Hg'),
        'symbol': 'inch Hg',
        'family': _('pressure')
    }, 'inch_Hg_60F': {
        'name': _('inch Hg 60F'),
        'symbol': 'inch Hg 60F',
        'family': _('pressure')
    }, 'international_british_thermal_unit': {
        'name': _('international Btu'),
        'symbol': 'international Btu',
        'family': _('energy')
    }, 'international_calorie': {
        'name': _('international calorie'),
        'symbol': 'international calorie',
        'family': _('energy')
    }, 'josephson_constant': {
        'name': _('Josephson constant'),
        'symbol': 'KJ',
        'family': _('constant'),
    }, 'joule': {
        'name': _('Joule'),
        'symbol': 'J',
        'family': _('energy'),
    }, 'jute': {
        'name': _('jute'),
        'symbol': 'jute',
        'family': _('volume'),
    }, 'katal': {
        'name': _('katal'),
        'symbol': 'kat',
        'family': _('catalytic activity'),
    }, 'kelvin': {
        'name': _('kelvin'),
        'symbol': 'K',
        'family': _('temperature'),
    }, 'kilometer_per_hour': {
        'name': _('kilometer per hour'),
        'symbol': 'km/hour',
        'family': _('speed'),
    }, 'kilometer_per_second': {
        'name': _('kilometer per second'),
        'symbol': 'km/second',
        'family': _('speed'),
    }, 'kip': {
        'name': _('kip'),
        'symbol': 'kip',
        'family': _('force'),
    }, 'kip_per_square_inch': {
        'name': _('kip per square inch'),
        'symbol': 'kip / sq inch',
        'family': _('pressure')
    }, 'knot': {
        'name': _('knot'),
        'symbol': 'kt',
        'family': _('speed')
    }, 'lambda': {
        'name': _('lambda'),
        'symbol': 'λ',
        'family': _('volume')
    }, 'lambert': {
        'name': _('lambert'),
        'symbol': 'L',
        'family': _('luminance')
    }, 'langley': {
        'name': _('langley'),
        'symbol': 'Ly',
        'family': _('radiation')
    }, 'lattice_spacing_of_Si': {
        'name': _('lattice constant'),
        'symbol': 'Ly',
        'family': _('constant')
    }, 'league': {
        'name': _('league'),
        'symbol': 'league',
        'obsolete': True,
        'family': _('length')
    }, 'leap_year': {
        'name': _('leap year'),
        'symbol': _('year'),
        'family': _('time')
    }, 'light_year': {
        'name': _('light year'),
        'symbol': _('ly'),
        'family': _('length')
    }, 'link': {
        'name': _('link'),
        'symbol': _('li'),
        'family': _('length')
    }, 'liter': {
        'name': _('liter'),
        'symbol': _('L'),
        'family': _('volume')
    }, 'ln10': {
        'name': _('log 10'),
        'symbol': _('ln10'),
        'family': _('dimesionless')
    }, 'long_hundredweight': {
        'name': _('long hundredweight'),
        'symbol': 'long hundredweight',
        'obsolete': True,
        'family': _('mass')
    }, 'long_ton': {
        'name': _('long ton'),
        'symbol': 'long ton',
        'obsolete': True,
        'family': _('length')
    }, 'lumen': {
        'name': _('lumen'),
        'symbol': 'lm',
        'family': _('luminous flux')
    }, 'lux': {
        'name': _('lux'),
        'symbol': 'lx',
        'family': _('illuminance')
    }, 'magnetic_flux_quantum': {
        'name': _('magnetic flux quantum'),
        'symbol': 'Φ',
        'family': _('magnetic flux quantum')
    }, 'maxwell': {
        'name': _('maxwell'),
        'symbol': 'Mx',
        'family': _('electrostatic unit of charge')
    }, 'mean_international_ampere': {
        'name': _('mean international ampere'),
        'symbol': 'mean international A',
        'family': _('electric current')
    }, 'mean_international_ohm': {
        'name': _('mean international Ohm'),
        'symbol': 'mean international Ω',
        'family': _('electric charge')
    }, 'mean_international_volt': {
        'name': _('mean international Volt'),
        'symbol': 'mean international V',
        'family': _('electric potential difference')
    }, 'mercury': {
        'name': _('density of mercury'),
        'symbol': '',
        'family': _('constant')
    }, 'mercury_60F': {
        'name': _('density of mercury 60F'),
        'symbol': '',
        'family': _('constant')
    }, 'meter': {
        'name': _('meter'),
        'symbol': 'm',
        'family': _('length')
    }, 'meter_per_second': {
        'name': _('meter per second'),
        'symbol': 'm/s',
        'family': _('speed')
    }, 'metric_horsepower': {
        'name': _('metric horsepower'),
        'symbol': 'hp',
        'family': _('power')
    }, 'metric_ton': {
        'name': _('metric ton'),
        'symbol': 't',
        'family': _('mass')
    }, 'micron': {
        'name': _('micron'),
        'symbol': 'μm',
        'family': _('length')
    }, 'mil': {
        'name': _('mil'),
        'symbol': 'ml',
        'family': _('volume')
    }, 'mile': {
        'name': _('mile'),
        'symbol': 'mile',
        'family': _('length')
    }, 'mile_per_hour': {
        'name': _('mile per hour'),
        'symbol': 'mph',
        'family': _('speed')
    }, 'millennium': {
        'name': _('millenium'),
        'symbol': _('millenium'),
        'family': _('time')
    }, 'milliarcsecond': {
        'name': _('milliarcsecond'),
        'symbol': 'mas',
        'family': _('angle')
    }, 'millimeter_Hg': {
        'name': _('Hg millimeter'),
        'symbol': 'mmHg',
        'family': _('pressure')
    }, 'minim': {
        'name': _('minim'),
        'symbol': 'min',
        'family': _('volume')
    }, 'minute': {
        'name': _('minute'),
        'symbol': 'min',
        'family': _('time')
    }, 'molar': {
        'name': _('molar'),
        'symbol': 'M',
        'family': _('concentration')
    }, 'molar_gas_constant': {
        'name': _('molar gas constant'),
        'symbol': 'R',
        'family': _('constant')
    }, 'mole': {
        'name': _('mole'),
        'symbol': 'mol',
        'family': _('amount of substance'),
    }, 'month': {
        'name': _('month'),
        'symbol': 'm',
        'family': _('time')
    }, 'nautical_mile': {
        'name': _('nautical mile'),
        'symbol': _('nm'),
        'family': _('length')
    }, 'neper': {
        'name': _('neper'),
        'symbol': _('Np'),
        'family': _('neper')
    }, 'neutron_mass': {
        'name': _('neutron mass'),
        'symbol': 'mN',
        'family': _('constant')
    }, 'newton': {
        'name': _('newton'),
        'symbol': 'N',
        'family': _('force')
    }, 'newtonian_constant_of_gravitation': {
        'name': _('newtonian constant of gravitation'),
        'symbol': 'N',
        'family': _('constant')
    }, 'nit': {
        'name': _('nit'),
        'symbol': 'cd/m²',
        'family': _('luminance')
    }, 'nuclear_magneton': {
        'name': _('nuclear magneton'),
        'symbol': 'μN',
        'family': _('magnetic moment')
    }, 'number_english': {
        'name': _('number english'),
        'symbol': 'm/g',
        'family': _('meter per gram')
    }, 'number_meter': {
        'name': _('number meter'),
        'symbol': 'm/g',
        'family': _('meter per gram')
    }, 'oersted': {
        'name': _('oersted'),
        'symbol': 'Oe',
        'family': _('H-field')
    }, 'ohm': {
        'name': _('ohm'),
        'symbol': 'Ω',
        'family': _('electric charge')
    }, 'oil_barrel': {
        'name': _('oil barrel'),
        'symbol': 'oil barrel',
        'family': 'volume'
    }, 'ounce': {
        'name': _('ounce'),
        'symbol': 'oz t',
        'family': _('mass')
    }, 'parsec': {
        'name': _('parsec'),
        'symbol': 'pc',
        'family': _('length')
    }, 'particle': {
        'name': _('particle'),
        'symbol': 'particle',
        'family': _('amount of substance')
    }, 'pascal': {
        'name': _('pascal'),
        'symbol': 'Pa',
        'family': _('pressure')
    }, 'peak_sun_hour': {
        'name': _('peak sun hour'),
        'symbol': 'peak sun hour',
        'family': _('time')
    }, 'peck': {
        'name': _('peck'),
        'symbol': 'peck',
        'family': _('volume')
    }, 'pennyweight': {
        'name': _('pennyweight'),
        'symbol': 'dwt',
        'family': _('mass')
    }, 'pi': {
        'name': 'pi',
        'symbol': 'π',
        'family': _('constant')
    }, 'pica': {
        'name': _('pica'),
        'symbol': 'pica',
        'family': _('length')
    }, 'pint': {
        'name': _('pint'),
        'symbol': 'pt',
        'family': _('volume')
    }, 'pixel': {
        'name': _('pixel'),
        'symbol': 'px',
        'family': _('unit')
    }, 'pixels_per_centimeter': {
        'name': _('pixel per centimeter'),
        'symbol': 'ppcm',
        'family': _('density of information')
    }, 'pixels_per_inch': {
        'name': _('pixel per inch'),
        'symbol': 'ppi',
        'family': _('density of information')
    }, 'planck_constant': {
        'name': _('Planck constant'),
        'symbol': 'h',
        'family': _('constant')
    }, 'planck_current': {
        'name': _('Planck current'),
        'symbol': _('qP'),
        'family': _('electric current')
    }, 'planck_length': {
        'name': _('Planck current'),
        'symbol': _('lP'),
        'family': _('length')
    }, 'planck_mass': {
        'name': _('Planck mass'),
        'symbol': _('mP'),
        'family': _('mass')
    }, 'planck_temperature': {
        'name': _('Planck temperature'),
        'symbol': _('TP'),
        'family': _('temperature')
    }, 'planck_time': {
        'name': _('Planck time'),
        'symbol': _('tP'),
        'family': _('time')
    }, 'point': {
        'name': _('point'),
        'symbol': 'pt',
        'family': _('length')
    }, 'poise': {
        'name': _('poise'),
        'symbol': 'P',
        'family': _('viscosity')
    }, 'pound': {
        'name': _('pound'),
        'symbol': 'lb',
        'family': _('mass')
    }, 'pound_force_per_square_inch': {
        'name': _('pound force per square inch'),
        'symbol': 'lb-force/inch²',
        'family': _('pressure')
    }, 'poundal': {
        'name': _('poundal'),
        'symbol': 'pdl',
        'family': _('force')
    }, 'proton_mass': {
        'name': _('proton mass'),
        'symbol': 'mP',
        'family': _('constant')
    }, 'quadrillion_Btu': {
        'name': _('quadrillion Btu'),
        'symbol': 'quadrillion Btu',
        'family': _('energy')
    }, 'quart': {
        'name': _('quart'),
        'symbol': 'qt',
        'family': _('volume')
    }, 'quarter': {
        'name': _('quarter'),
        'symbol': 'qr',
        'family': _('length')
    }, 'radian': {
        'name': _('radian'),
        'symbol': 'rad',
        'family': _('angle')
    }, 'rads': {
        'name': _('rads'),
        'symbol': 'rads',
        'family': _('absorption')
    }, 'reciprocal_centimeter': {
        'name': _('reciprocal centimeter'),
        'symbol': 'cm⁻¹',
        'family': _('reciprocal length')
    }, 'refrigeration_ton': {
        'name': _('refrigeration ton'),
        'symbol': 'TR',
        'family': _('power')
    }, 'rem': {
        'name': _('roentgen equivalent man'),
        'symbol': 'rem',
        'family': _('absorption')
    }, 'revolutions_per_minute': {
        'name': _('revolutions per minute'),
        'symbol': _('rpm'),
        'family': _('angular speed')
    }, 'revolutions_per_second': {
        'name': _('revolutions per second'),
        'symbol': _('rps'),
        'family': _('angular speed')
    }, 'reyn': {
        'name': _('reyn'),
        'symbol': _('reyn'),
        'family': _('viscosity')
    }, 'rhe': {
        'name': _('reciprocal poise'),
        'symbol': 'P⁻¹',
        'family': _('reciprocal viscosity')
    }, 'rod': {
        'name': _('rod'),
        'symbol': 'rod',
        'family': _('length')
    }, 'roentgen': {
        'name': _('roentgen'),
        'symbol': 'R',
        'family': _('radiation exposure')
    }, 'rutherford': {
        'name': _('rutherford'),
        'symbol': 'Rd',
        'family': _('frequency')
    }, 'rydberg': {
        'name': _('rydberg'),
        'symbol': 'Ry',
        'family': _('energy')
    }, 'rydberg_constant': {
        'name': _('rydberg constant'),
        'symbol': 'R∞',
        'family': _('constant')
    }, 'scaled_point': {
        'name': _('scaled point'),
        'symbol': 'Pt',
        'family': _('length')
    }, 'scruple': {
        'name': _('scruple'),
        'symbol': '℈',
        'family': _('volume'),
    }, 'second': {
        'name': _('second'),
        'symbol': 's',
        'family': _('time'),
    }, 'second_radiation_constant': {
        'name': _('second radiation constant'),
        'symbol': 'c2',
        'family': _('constant'),
    }, 'shake': {
        'name': _('shake'),
        'symbol': 'shake',
        'family': _('time'),
    }, 'shot': {
        'name': _('shot'),
        'symbol': 'shot',
        'family': _('length'),
    }, 'sidereal_day': {
        'name': _('sidereal day'),
        'symbol': pgettext_lazy('unit', 'sidereal day'),
        'family': _('time')
    }, 'sidereal_month': {
        'name': _('sidereal month'),
        'symbol': pgettext_lazy('unit', 'sidereal month'),
        'family': _('time')
    }, 'sidereal_year': {
        'name': _('sidereal year'),
        'symbol': pgettext_lazy('unit', 'sidereal year'),
        'family': _('time')
    }, 'siemens': {
        'name': _('siemens'),
        'symbol': 'S',
        'family': _('electrical conductance')
    }, 'sievert': {
        'name': _('sievert'),
        'symbol': 'Sv',
        'family': _('absorption')
    }, 'slinch': {
        'name': _('slinch'),
        'symbol': 'slinch',
        'family': _('mass')
    }, 'slug': {
        'name': _('slug'),
        'symbol': 'slug',
        'family': _('mass')
    }, 'speed_of_light': {
        'name': _('speed of light'),
        'symbol': 'c',
        'family': _('speed')
    }, 'square_degree': {
        'name': _('square degree'),
        'symbol': '(°)²',
        'family': _('solid angle')
    }, 'square_foot': {
        'name': _('square foot'),
        'symbol': 'sq ft',
        'family': _('surface')
    }, 'square_inch': {
        'name': _('square inch'),
        'symbol': 'sq in',
        'family': _('surface')
    }, 'square_league': {
        'name': _('square league'),
        'symbol': 'sq league',
        'family': _('surface')
    }, 'square_mile': {
        'name': _('square mile'),
        'symbol': 'sq mile',
        'family': _('surface')
    }, 'square_rod': {
        'name': _('square rod'),
        'symbol': 'sq rod',
        'family': _('surface')
    }, 'square_survey_mile': {
        'name': _('square survey mile'),
        'symbol': 'sq survey mile',
        'family': _('surface')
    }, 'square_yard': {
        'name': _('square yard'),
        'symbol': 'sq yd',
        'family': _('surface')
    }, 'standard_atmosphere': {
        'name': _('standard atmosphere'),
        'symbol': 'atm',
        'family': _('pressure')
    }, 'standard_gravity': {
        'name': _('standard gravity'),
        'symbol': 'g0',
        'family': _('constant')
    }, 'standard_liter_per_minute': {
        'name': _('standard liter per minute'),
        'symbol': 'SLM',
        'family': _('volumetric flow rate')
    }, 'statampere': {
        'name': 'statampere',
        'symbol': 'statA',
        'family': _('electric current')
    }, 'statfarad': {
        'name': 'statfarad',
        'symbol': 'statF',
        'family': _('electrical capacitance')
    }, 'stathenry': {
        'name': 'stathenry',
        'symbol': 'stathenry',
        'family': _('electrical inductance'),
    }, 'statmho': {
        'name': 'statmho',
        'symbol': 'statmho',
        'family': _('electrical conductance')
    }, 'statohm': {
        'name': 'statohm',
        'symbol': 'statohm',
        'family': _('electrical resistance')
    }, 'stattesla': {
        'name': 'stattesla',
        'symbol': 'stattesla',
        'family': _('magnetic field')
    }, 'statvolt': {
        'name': 'statvolt',
        'symbol': 'statvolt',
        'family': _('electric potential difference')
    }, 'statweber': {
        'name': 'statweber',
        'symbol': 'statweber',
        'family': _('flux density')
    }, 'stefan_boltzmann_constant': {
        'name': _('Stefan-Boltzmann constant'),
        'symbol': 'σ',
        'family': _('constant')
    }, 'steradian': {
        'name': _('steradian'),
        'symbol': 'sr',
        'family': _('solid angle')
    }, 'stere': {
        'name': _('stere'),
        'symbol': 'st',
        'family': _('volume')
    }, 'stilb': {
        'name': _('stilb'),
        'symbol': 'sb',
        'family': _('luminance')
    }, 'stokes': {
        'name': _('stokes'),
        'symbol': 'St',
        'family': _('viscosity')
    }, 'stone': {
        'name': _('stone'),
        'symbol': 'st.',
        'family': _('mass')
    }, 'survey_foot': {
        'name': _('survey foot'),
        'symbol': 'survey ft',
        'family': _('length')
    }, 'survey_mile': {
        'name': _('survey mile'),
        'symbol': 'survey mile',
        'family': _('length')
    }, 'svedberg': {
        'name': 'svedberg',
        'symbol': 'S',
        'family': _('time')
    }, 'synodic_month': {
        'name': _('synodic month'),
        'symbol': _('synodic month'),
        'family': _('time')
    }, 'tablespoon': {
        'name': _('tablespoon'),
        'symbol': _('tablespoon'),
        'family': _('volume')
    }, 'tansec': {
        'name': 'tansec',
        'symbol': 'tansec',
        'family': _('constant')
    }, 'teaspoon': {
        'name': _('teaspoon'),
        'symbol': _('teaspoon'),
        'family': _('volume')
    }, 'technical_atmosphere': {
        'name': _('technical atmosphere'),
        'symbol': 'at',
        'family': _('pressure')
    }, 'tesla': {
        'name': 'tesla',
        'symbol': 'T',
        'family': _('magnetic field')
    }, 'tex': {
        'name': 'tex',
        'symbol': 'tex',
        'family': _('linear mass density')
    }, 'tex_cicero': {
        'name': 'tex cicero',
        'symbol': 'tex cicero',
        'family': _('length')
    }, 'tex_didot': {
        'name': 'tex didot',
        'symbol': 'tex didot',
        'family': _('length')
    }, 'tex_pica': {
        'name': 'tex pica',
        'symbol': 'tex pica',
        'family': _('length')
    }, 'tex_point': {
        'name': 'tex point',
        'symbol': 'tex point',
        'family': _('length')
    }, 'therm': {
        'name': _('therm'),
        'symbol': 'Thm',
        'family': _('energy')
    }, 'thermochemical_british_thermal_unit': {
        'name': _('thermochemical Btu'),
        'symbol': 'thermochemical Btu',
        'family': _('energy')
    }, 'thomson_cross_section': {
        'name': _('Thomson cross section'),
        'symbol': 'Thomson cross section',
        'family': _('surface')
    }, 'thou': {
        'name': 'thou',
        'symbol': 'thou',
        'family': _('length')
    }, 'ton': {
        'name': _('ton'),
        'symbol': 't',
        'family': _('mass')
    }, 'ton_TNT': {
        'name': _('ton of TNT'),
        'symbol': 'ton of TNT',
        'family': _('energy')
    }, 'tonne_of_oil_equivalent': {
        'name': _('ton of oil equivalent'),
        'symbol': _('toe'),
        'family': _('energy')
    }, 'torr': {
        'name': 'torr',
        'symbol': 'torr',
        'family': _('pressure')
    }, 'tropical_month': {
        'name': _('tropical month'),
        'symbol': _('tropical month'),
        'family': _('time')
    }, 'tropical_year': {
        'name': _('tropical year'),
        'symbol': _('tropical year'),
        'family': _('time')
    }, 'troy_ounce': {
        'name': _('troy ounce'),
        'symbol': _('troy ounce'),
        'family': _('mass')
    }, 'troy_pound': {
        'name': _('troy pound'),
        'symbol': _('troy pound'),
        'family': _('mass')
    }, 'turn': {
        'name': _('turn'),
        'symbol': _('turn'),
        'family': _('angle')
    }, 'unified_atomic_mass_unit': {
        'name': _('unified atomic mass unit'),
        'symbol': _('unified atomic mass unit'),
        'family': _('mass')
    }, 'unit_pole': {
        'name': _('unit pole'),
        'symbol': 'pole',
        'family': _('length')
    }, 'vacuum_permeability': {
        'name': _('vacuum permeability'),
        'symbol': 'μ0',
        'family': _('constant')
    }, 'vacuum_permittivity': {
        'name': _('vacuum permittivity'),
        'symbol': 'ε0',
        'family': _('constant')
    }, 'volt': {
        'name': 'Volt',
        'symbol': 'V',
        'family': _('electric potential difference')
    }, 'volt_ampere': {
        'name': 'Volt Ampere',
        'symbol': 'VA',
        'family': _('power')
    }, 'von_klitzing_constant': {
        'name': _('Von Klitzing constant'),
        'symbol': 'RK',
        'family': _('constant'),
    }, 'water': {
        'name': _('Water volumic mass'),
        'symbol': 'g/m³',
        'family': _('constant'),
    }, 'water_39F': {
        'name': _('Water volumic mass at 39F'),
        'symbol': 'g/m³',
        'family': _('constant'),
    }, 'water_60F': {
        'name': _('Water volumic mass at 60F'),
        'symbol': 'g/m³',
        'family': _('constant'),
    }, 'watt': {
        'name': 'Watt',
        'symbol': 'W',
        'family': _('power'),
    }, 'watt_hour': {
        'name': 'Watt',
        'symbol': 'Wh',
        'family': _('energy'),
    }, 'weber': {
        'name': 'weber',
        'symbol': 'weber',
        'family': _('flux density')
    }, 'week': {
        'name': 'week',
        'symbol': 'w',
        'family': _('time'),
    }, 'wien_frequency_displacement_law_constant': {
        'name': _('wien frequency displacement law constant'),
        'symbol': 'vpeak',
        'family': _('constant')
    }, 'wien_u': {
        'name': _('wien u'),
        'symbol': '',
        'family': _('constant')
    }, 'wien_wavelength_displacement_law_constant': {
        'name': _('wien displacement law constant'),
        'symbol': 'λpeak',
        'family': _('constant')
    }, 'wien_x': {
        'name': _('wien x'),
        'symbol': '',
        'family': _('constant')
    }, 'x_unit_Cu': {
        'name': _('Cu X-ray wavelength'),
        'symbol': '',
        'family': _('constant')
    }, 'x_unit_Mo': {
        'name': _('Mo X-ray wavelength'),
        'symbol': '',
        'family': _('constant')
    }, 'yard': {
        'name': 'yard',
        'symbol': 'yd',
        'family': _('length')
    }, 'year': {
        'name': _('year'),
        'symbol': _('year'),
        'family': _('time')
    }, 'zeta': {
        'name': 'zeta',
        'symbol': '',
        'family': _('constant')
    }

}



from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from django.views.decorators.vary import vary_on_cookie
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.status import HTTP_404_NOT_FOUND, HTTP_400_BAD_REQUEST
from rest_framework.views import APIView
from rest_framework.viewsets import ViewSet

from .models import UnitSystem, UnitConverter
from .serializers import UnitSerializer, UnitSystemListSerializer, UnitSystemDetailSerializer, \
    UnitConversionPayloadSerializer
from geocurrency.converters.serializers import ConverterResultSerializer


class UnitSystemViewset(ViewSet):
    """
    View for currency
    """
    lookup_field = 'system_name'

    language_header = openapi.Parameter('Accept-Language', openapi.IN_HEADER, description="language",
                                        type=openapi.TYPE_STRING)
    language = openapi.Parameter('language', openapi.IN_QUERY, description="language",
                                 type=openapi.TYPE_STRING)

    unit_systems_response = openapi.Response('List of unit systems', UnitSystemListSerializer)
    unit_system_response = openapi.Response('Dimensions and units in a system', UnitSystemDetailSerializer)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(manual_parameters=[language, language_header], responses={200: unit_systems_response})
    def list(self, request):
        language = request.GET.get('language', request.LANGUAGE_CODE)
        us = UnitSystem(fmt_locale=language)
        us = [{'system_name': s} for s in us.available_systems()]
        return Response(us, content_type="application/json")

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(manual_parameters=[language, language_header], responses={200: unit_system_response})
    def retrieve(self, request, system_name):
        language = request.GET.get('language', request.LANGUAGE_CODE)
        try:
            us = UnitSystem(system_name=system_name, fmt_locale=language)
            serializer = UnitSystemDetailSerializer(us, context={'request': request})
            return Response(serializer.data, content_type="application/json")
        except (ValueError, KeyError) as e:
            return Response("Unknown unit system: " + str(e), status=HTTP_404_NOT_FOUND)


class UnitViewset(ViewSet):
    """
    View for currency
    """
    lookup_field = 'unit_name'
    language_header = openapi.Parameter('Accept-Language', openapi.IN_HEADER, description="language",
                                        type=openapi.TYPE_STRING)
    language = openapi.Parameter('language', openapi.IN_QUERY, description="language",
                                 type=openapi.TYPE_STRING)
    family = openapi.Parameter('family', openapi.IN_QUERY, description="Unit dimension",
                                 type=openapi.TYPE_STRING)
    units_response = openapi.Response('List of units in a system', UnitSerializer)
    unit_response = openapi.Response('Detail of a unit', UnitSerializer)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(manual_parameters=[family, language, language_header], responses={200: units_response})
    def list(self, request, system_name):
        language = request.GET.get('language', request.LANGUAGE_CODE)
        try:
            us = UnitSystem(system_name=system_name, fmt_locale=language)
            if dimension := request.GET.get('family'):
                available_units = [unit.code for unit in us.units_per_family().get(dimension)]
            else:
                available_units = us.available_unit_names()
            units = []
            if available_units:
                units = [us.unit(unit_name=unit_name) for unit_name in available_units]
            serializer = UnitSerializer(units, many=True, context={'request': request})
            return Response(serializer.data)
        except KeyError:
            return Response('Invalid Unit System', status=status.HTTP_404_NOT_FOUND)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(manual_parameters=[language, language_header], responses={200: unit_response})
    def retrieve(self, request, system_name, unit_name):
        """
        Get unit information for unit in unit system
        """
        language = request.GET.get('language', request.LANGUAGE_CODE)
        try:
            us = UnitSystem(system_name=system_name, fmt_locale=language)
            unit = us.unit(unit_name=unit_name)
            serializer = UnitSerializer(unit, context={'request': request})
            return Response(serializer.data, content_type="application/json")
        except (ValueError, KeyError):
            return Response("Unknown unit", status=HTTP_404_NOT_FOUND)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(manual_parameters=[language, language_header], responses={200: units_response})
    @action(methods=['GET'], detail=True, url_path='compatible', url_name='compatible_units')
    def compatible_units(self, request, system_name, unit_name):
        language = request.GET.get('language', request.LANGUAGE_CODE)
        try:
            us = UnitSystem(system_name=system_name, fmt_locale=language)
            unit = us.unit(unit_name=unit_name)
            compatible_units = [us.unit(unit_name=cunit) for cunit in map(str, unit.unit.compatible_units())]
            serializer = UnitSerializer(compatible_units, many=True, context={'request': request})
            return Response(serializer.data, content_type="application/json")
        except (ValueError, KeyError):
            return Response("Unknown unit", status=HTTP_404_NOT_FOUND)


class ConvertView(APIView):

    @swagger_auto_schema(request_body=UnitConversionPayloadSerializer,
                         responses={200: ConverterResultSerializer})
    @action(['POST'], detail=False, url_path='', url_name="convert")
    def post(self, request, *args, **kwargs):
        """
        Converts a list of amounts with currency and date to a reference currency
        :param request: HTTP request
        """
        cps = UnitConversionPayloadSerializer(data=request.data)
        if not cps.is_valid():
            return Response(cps.errors, status=HTTP_400_BAD_REQUEST, content_type="application/json")
        cp = cps.create(cps.validated_data)
        try:
            converter = UnitConverter.load(cp.batch_id)
        except KeyError:
            converter = UnitConverter(
                id=cp.batch_id,
                base_system=cp.base_system,
                base_unit=cp.base_unit
            )
        if cp.data:
            if errors := converter.add_data(data=cp.data):
                return Response(errors, status=HTTP_400_BAD_REQUEST)
        if cp.eob or not cp.batch_id:
            result = converter.convert()
            serializer = ConverterResultSerializer(result)
            return Response(serializer.data, content_type="application/json")
        else:
            return Response({'id': converter.id, 'status': converter.status}, content_type="application/json")



import logging
from datetime import date

import pint
from django.utils.translation import ugettext as _
from geocurrency.converters.models import BaseConverter, ConverterResult, ConverterResultDetail, ConverterResultError

from . import UNIT_EXTENDED_DEFINITION


class Amount:
    system = None
    unit = None
    value = 0
    date_obj = None

    def __init__(self, system: str, unit: str, value: float, date_obj: date = None):
        self.system = system
        self.unit = unit
        self.value = value
        self.date_obj = date_obj

    def __repr__(self):
        return f'{self.value} {self.unit} ({self.system})'


class Unit:
    pass


class UnitFamily:
    family = None
    units = []

    def __init__(self, family: str, units: [Unit]):
        self.family = family
        self.units = units


class UnitSystem:
    ureg = None
    system_name = None
    system = None

    def __init__(self, system_name='SI', fmt_locale='en'):
        found = False
        for available_system in UnitSystem.available_systems():
            if system_name.lower() == available_system.lower():
                system_name = available_system
                found = True
        if not found:
            raise ValueError("Invalid unit system")
        self.system_name = system_name
        try:
            self.ureg = pint.UnitRegistry(system=system_name, fmt_locale=fmt_locale)
            self.system = getattr(self.ureg.sys, system_name)
        except (FileNotFoundError, AttributeError):
            raise ValueError("Invalid unit system")

    @classmethod
    def available_systems(cls) -> [str]:
        """
        List of available Unit Systems
        :return: Array of string
        """
        ureg = pint.UnitRegistry(system='SI')
        return dir(ureg.sys)

    @classmethod
    def is_valid(cls, system: str) -> bool:
        us = cls()
        return system in us.available_systems()

    def current_system(self) -> pint.UnitRegistry:
        """
        Return current pint.UnitRegistry
        """
        return self.ureg

    def unit(self, unit_name):
        return Unit(unit_system=self, code=unit_name)

    @property
    def units(self):
        """
        Wrapper used by serializer
        """
        return [UnitFamily(family=family, units=units) for family, units in self.units_per_family().items()]

    def available_unit_names(self) -> [str]:
        """
        List of available units for a given Unit system
        :return: Array of names of Unit systems
        """
        return dir(self.system)

    def unit_dimensionality(self, unit: str) -> str:
        """
        User friendly representation of the dimension
        :param unit: name of the unit to display
        :return: Human readable dimension
        """
        return Unit.dimensionality_string(unit_system=self.system, unit_str=unit)

    def units_per_family(self) -> []:
        """
        List of units per dimension.
        Dimension is the dimension as identified in Units.UNIT_EXTENDED_DEFINITION
        :return: array of dict of dimensions, with lists of unit strings
        """
        units_array = self.available_unit_names()
        output = {}
        for unit_str in units_array:
            try:
                unit = self.unit(unit_name=unit_str)
            except ValueError:
                continue
            try:
                output[str(unit.family)].append(unit)
            except KeyError:
                output[str(unit.family)] = [unit]
        return output

    def units_per_dimensionality(self) -> {}:
        """
        List of units per dimension
        :return: dict of dimensions, with lists of unit strings
        """
        units_array = self.available_unit_names()
        output = {}
        for unit_str in units_array:
            dimension = Unit.dimensionality_string(self, unit_str)
            try:
                output[dimension].append(unit_str)
            except KeyError:
                output[dimension] = [unit_str]
        return output

    @property
    def dimensionalities(self) -> [str]:
        """
        List of dimensions available in the Unit system
        :return: list of dimensions for Unit system
        """
        return set([Unit.dimensionality_string(self, unit_str) for unit_str in dir(self.system)])


class Unit:
    unit_system = None
    code = None
    unit = None

    def __init__(self, unit_system: UnitSystem, code: str, ):
        """
        :param unit_system: UnitSystem instance
        :param code: code of the pint.Unit
        """
        self.unit_system = unit_system
        self.code = code
        try:
            self.unit = getattr(unit_system.system, code)
        except pint.errors.UndefinedUnitError:
            raise ValueError("invalid unit for system")

    @classmethod
    def is_valid(self, name):
        us_si = UnitSystem(system_name='SI')
        all_units = []
        for us in us_si.available_systems():
            all_units.extend(dir(getattr(us_si.ureg.sys, us)))
        return name in set(all_units)

    @property
    def family(self):
        return self.unit_family(self.code)

    @property
    def name(self):
        return self.unit_name(self.code)

    @property
    def symbol(self):
        return self.unit_symbol(self.code)

    @property
    def dimension(self):
        return self.unit.dimensionality

    @staticmethod
    def unit_family(unit_str: str) -> str:
        try:
            ext_unit = UNIT_EXTENDED_DEFINITION.get(unit_str)
            return ext_unit['family']
        except (KeyError, TypeError) as e:
            logging.error(f'No UNIT_EXTENDED_DEFINITION for unit {unit_str}')
            return ''

    @staticmethod
    def unit_name(unit_str: str) -> str:
        try:
            ext_unit = UNIT_EXTENDED_DEFINITION.get(unit_str)
            return ext_unit['name']
        except (KeyError, TypeError) as e:
            logging.error(f'No UNIT_EXTENDED_DEFINITION for unit {unit_str}')
            return unit_str

    @staticmethod
    def unit_symbol(unit_str: str) -> str:
        try:
            ext_unit = UNIT_EXTENDED_DEFINITION.get(unit_str)
            return ext_unit['symbol']
        except (KeyError, TypeError) as e:
            logging.error(f'No UNIT_EXTENDED_DEFINITION for unit {unit_str}')
            return ''

    @staticmethod
    def dimensionality_string(unit_system: UnitSystem, unit_str: str) -> str:
        """
        Converts pint dimensionality string to human readable string
        :param unit_system: UnitSystem
        :param unit_str: Unit name
        :return: str
        """
        ds = str(getattr(unit_system.ureg, unit_str).dimensionality).replace('[', '').replace(']', '')
        ds = ds.replace(' ** ', '^')
        ds = ds.split()
        return ' '.join([_(d) for d in ds])

    @staticmethod
    def translated_name(unit_system: UnitSystem, unit_str: str) -> str:
        return '{}'.format(unit_system.ureg[unit_str])

    @property
    def readable_dimension(self):
        """
        Wrapper around Unit.dimensionality_string
        """
        return Unit.dimensionality_string(unit_system=self.unit_system, unit_str=self.code)


class UnitConverter:
    pass


class UnitConverter(BaseConverter):
    base_system = None
    base_unit = None

    def __init__(self, base_system: str, base_unit: str, id: str = None):
        super(UnitConverter, self).__init__(id=id)
        self.base_system = base_system
        self.base_unit = base_unit
        self.system = UnitSystem(system_name=base_system)
        self.unit = Unit(unit_system=self.system, code=base_unit)
        self.compatible_units = [str(u) for u in self.unit.unit.compatible_units()]

    def add_data(self, data: []) -> []:
        """
        Check data and add it to the dataset
        Return list of errors
        """
        errors = super(UnitConverter, self).add_data(data)
        return errors

    def check_data(self, data):
        """
        Validates that the data contains
        - currency (str)
        - amount (float)
        - date (YYYY-MM-DD)
        """
        from .serializers import UnitAmountSerializer
        errors = []
        for line in data:
            serializer = UnitAmountSerializer(data=line)
            if serializer.is_valid():
                self.data.append(serializer.create(serializer.validated_data))
            else:
                errors.append(serializer.errors)
        return errors

    @classmethod
    def load(cls, id: str) -> UnitConverter:
        uc = super(UnitConverter, cls).load(id)
        uc.system = UnitSystem(system_name=uc.base_system)
        uc.unit = Unit(unit_system=uc.system, code=uc.base_unit)
        return uc

    def save(self):
        system = self.system
        unit = self.unit
        self.system = None
        self.unit = None
        super(UnitConverter, self).save()
        self.system = system
        self.unit = unit

    def convert(self) -> ConverterResult:
        """
        Converts data to base unit in base system
        """

        result = ConverterResult(id=self.id, target=self.base_unit)
        us = UnitSystem(system_name=self.base_system)
        Q_ = us.ureg.Quantity
        for amount in self.data:
            if amount.unit not in self.compatible_units:
                error = ConverterResultError(
                    unit=amount.unit,
                    original_value=amount.value,
                    date=amount.date_obj,
                    error=_('Incompatible units')
                )
                result.errors.append(error)
                continue
            try:
                quantity = Q_(amount.value, amount.unit)
                out = quantity.to(self.base_unit)
                result.increment_sum(out.magnitude)
                detail = ConverterResultDetail(
                    unit=amount.unit,
                    original_value=amount.value,
                    date=amount.date_obj,
                    conversion_rate=0,
                    converted_value=out.magnitude
                )
                result.detail.append(detail)
            except pint.UndefinedUnitError:
                error = ConverterResultError(
                    unit=amount.currency,
                    original_value=amount.amount,
                    date=amount.date_obj,
                    error=_('Undefined unit in the registry')
                )
                result.errors.append(error)
        self.end_batch(result.end_batch())
        return result


class UnitConversionPayload:
    data = None
    base_system = ''
    base_unit = ''
    key = ''
    batch_id = ''
    eob = False

    def __init__(self, base_system, base_unit, data=None, key=None, batch_id=None, eob=False):
        self.data = data
        self.base_system = base_system
        self.base_unit = base_unit
        self.key = key
        self.batch_id = batch_id
        self.eob = eob



from django.db import models
from django.db.models import OuterRef, Subquery
from django_filters import rest_framework as filters

from .models import Rate


class RateFilter(filters.FilterSet):
    user = filters.BooleanFilter(label="filter rate associated to connected user", method='user_filter')
    key = filters.CharFilter(label="filter rates with key", method='key_filter')
    value_date = filters.DateFilter(label="filter rates at a specific date",
                                    field_name='value_date', lookup_expr='exact')
    from_obj = filters.DateFilter(label="filter rates after a specific date (included)",
                                  field_name='value_date', lookup_expr='gte')
    to_obj = filters.DateFilter(label="filter rates before a specific date (included)",
                                field_name='value_date', lookup_expr='lte')
    value = filters.NumberFilter(label="filter rates with a specific value", field_name='value', lookup_expr='exact')
    lower_bound = filters.NumberFilter(label="filter rates with a value higher than the given value",
                                       field_name='value', lookup_expr='gte')
    higher_bound = filters.NumberFilter(label="filter rates with a value lower than the given value",
                                        field_name='value', lookup_expr='lte')
    currency = filters.CharFilter(label="filter by target currency", field_name='currency', lookup_expr='iexact')
    base_currency = filters.CharFilter(label="filter by base currency", field_name='base_currency',
                                       lookup_expr='iexact')
    currency_latest_values = filters.CharFilter(label="Only output latest rates for currency",
                                                method='currency_latest_values_filter')
    base_currency_latest_values = filters.CharFilter(label="Only output latest rates for currency",
                                                     method='base_currency_latest_values_filter')

    class Meta:
        model = Rate
        fields = [
            'user', 'key',
            'value_date', 'from_obj', 'to_obj',
            'value', 'lower_bound', 'higher_bound',
            'currency', 'base_currency'

        ]

    def user_filter(self, queryset, name, value):
        if self.request and self.request.user and self.request.user.is_authenticated:
            return queryset.filter(**{
                'user': self.request.user,
            })

    def key_filter(self, queryset, name, value):
        if self.request and self.request.user and self.request.user.is_authenticated:
            return queryset.filter(**{
                'user': self.request.user,
                'key': value
            })

    def currency_latest_values_filter(self, queryset, name, value):
        queryset = queryset.filter(currency=value)
        latest = queryset.filter(currency=OuterRef('currency')).order_by('-value_date')
        return queryset.annotate(
            currency_latest=Subquery(latest.values('value_date')[:1])
        ).filter(value_date=models.F('currency_latest'))

    def base_currency_latest_values_filter(self, queryset, name, value):
        queryset = queryset.filter(base_currency=value)
        latest = queryset.filter(base_currency=OuterRef('base_currency')).order_by('-value_date')
        return queryset.annotate(
            base_currency_latest=Subquery(latest.values('value_date')[:1])
        ).filter(value_date=models.F('base_currency_latest'))



import os
import pytz
from pytz import timezone
import re
import requests
from datetime import datetime
from pycountry import countries
from countryinfo import CountryInfo

from django.core.cache import cache
from django.conf import settings
from django.db import models


from .helpers import ColorProximity, hextorgb
from .settings import *


class CountryNotFoundError(Exception):
    msg = 'Country not found'


class CountryManager(models.Manager):

    def get_by_color(self, color, proximity=1):
        """
        Take a hex color and finds near country based on flag and color proximity
        :param color: hex color (FFF, FFFFFF, FFFFFFFF, #FFF, #FFFFFF, #FFFFFFFF)
        :param proximity: succes rate, positive if below (100 is opposite, 0 is identical
        """
        cp = ColorProximity()
        rgb_color = hextorgb(color)
        countries = []
        for country in Country.objects.filter(colors__isnull=False):
            for fc in country.colors.split(','):
                flag_color = hextorgb(fc)
                if cp.proximity(rgb_color, flag_color) < proximity:
                    countries.append(country.pk)
                    break
        return Country.objects.filter(pk__in=set(countries))


class Country:
    # data extracted from pycountry as basic data if CountryInfo for this country does not exist
    alpha_2 = None
    alpha_3 = None
    name = None
    numeric = None

    def __init__(self, alpha_2):
        """
        Init a Country object with an alpha2 code
        :params country_name: ISO-3166 alpha_2 code
        """
        country = countries.get(alpha_2=alpha_2)
        if not country:
            raise CountryNotFoundError("Invalid country alpha2 code")
        self.alpha_2 = country.alpha_2
        self.alpha_3 = country.alpha_3
        self.name = country.name
        self.numeric = country.numeric

    @classmethod
    def all_countries(cls):
        """
        List all countries, instanciate CountryInfo for each country in pycountry.countries
        """
        return list(map(lambda x: cls(x.alpha_2), countries))

    def base(self):
        """
        Returns a basic representation of a country with name and iso codes
        """
        return countries.get(alpha_2=self.alpha_2)._fields

    def currencies(self):
        ci = CountryInfo(self.alpha_2)
        return ci.currencies()

    def info(self):
        return CountryInfo(self.alpha_2).info()

    @property
    def unit_system(self):
        if self.alpha_2 == 'US':
            return 'US'
        if self.alpha_2 == 'LR':
            return 'US'
        if self.alpha_2 == 'MM':
            return 'imperial'
        return 'SI'

    @property
    def timezones(self):
        output = []
        fmt = '%z'
        base_time = datetime.utcnow()
        for tz_info in pytz.country_timezones[self.alpha_2]:
            tz = timezone(tz_info)
            offset = tz.localize(base_time).strftime(fmt)
            numeric_offset = float(offset[:-2] + '.' + offset[-2:])
            output.append({
                'name': tz_info,
                'offset': f'UTC {offset}',
                'numeric_offset': numeric_offset,
                'current_time': base_time.astimezone(tz).strftime('%Y-%m-%d %H:%M')
            })
        return sorted(output, key=lambda x: x['numeric_offset'])

    @property
    def flag_path(self):
        """
        Path to the flag temporary file
        :return: string, absolute path to the flag file
        """
        return os.path.join(settings.MEDIA_ROOT, self.alpha_2 + '.svg')

    def flag_exists(self):
        """
        Checks if flag file exists
        :return: bool, True if flag exists, False otherwise
        """
        return os.path.exists(self.flag_path)

    def download_flag(self):
        """
        Downloads flag for country in temporary path
        :return: Path to the file
        """
        if not self.flag_exists():
            response = requests.get(FLAG_SOURCE.format(alpha_2=self.alpha_2))
            try:
                flag_content = response.text
                flag_file = open(self.flag_path, 'w')
                flag_file.write(flag_content)
                flag_file.close()
                return self.flag_path
            except IOError:
                print("unable to write file", self.flag_path)
                return None

    def analyze_flag(self):
        """
        Analyze colors of the flag for the country and caches the result
        :returns: array, list of colors
        """
        flag_path = os.path.join(settings.MEDIA_ROOT, self.alpha_2 + '.svg')
        # Checks if flag has been downloaded, downloads it otherwise, and return None if download failed
        if not self.flag_exists() and not self.download_flag():
            return None
        with open(flag_path, 'r') as flag:
            content = flag.read()
            result = re.findall(r'\#[0-9A-Fa-f]{1,2}[0-9A-Fa-f]{1,2}[0-9A-Fa-f]{1,2}', content)
            if result:
                cache.set('COLORS-' + self.alpha_2, result)
            return result

    def colors(self):
        """
        List colors present in the flag of the country
        :returns: array, list of colors
        """
        if colors := cache.get('COLORS-' + self.alpha_2):
            return colors
        else:
            return self.analyze_flag()

    @property
    def region(self) -> str:
        return CountryInfo(self.alpha_2).region()

    @property
    def subregion(self) -> str:
        return CountryInfo(self.alpha_2).subregion()

    @property
    def tld(self) -> str:
        return CountryInfo(self.alpha_2).tld()

    @property
    def capital(self) -> str:
        return CountryInfo(self.alpha_2).capital()



import logging
from datetime import date
from typing import Iterator

from django.contrib.auth.models import User
from django.core.cache import cache
from django.db import models
from iso4217 import Currency as Iso4217

from geocurrency.countries.models import Country
from . import CURRENCY_SYMBOLS, DEFAULT_SYMBOL


class CurrencyNotFoundError(Exception):
    msg = 'Currency not found'


class Currency:
    pass


class Currency:
    code = None
    name = None
    currency_name = None
    exponent = None
    number = 0
    value = None

    def __init__(self, code):
        """
        Returns a iso4217.Currency instance
        """
        try:
            i = Iso4217(code)
            for a in ['code', 'name', 'currency_name', 'exponent', 'number', 'value']:
                setattr(self, a, getattr(i, a))
        except ValueError:
            raise CurrencyNotFoundError('Invalid currency code')

    @classmethod
    def is_valid(cls, cur: str) -> bool:
        """
        Checks if currency is part of iso4217
        """
        try:
            Currency(cur)
            return True
        except CurrencyNotFoundError:
            return False

    @classmethod
    def all_currencies(cls) -> Iterator[Currency]:
        """
        Returns an array of currencies
        """
        return [Currency(c.code) for c in Iso4217]

    @property
    def countries(self) -> Iterator[Country]:
        countries = []
        if cached_countries := cache.get(self.code + 'COUNTRIES'):
            return [Country(alpha_2) for alpha_2 in cached_countries]
        for country in Country.all_countries():
            try:
                if self.code in country.currencies():
                    countries.append(country)
            except KeyError:
                # Some countries listed in pycountry are not present in countryinfo
                pass
        cache.set(self.code + 'COUNTRIES', [country.alpha_2 for country in countries])
        return countries

    @classmethod
    def get_for_country(cls, alpha2: str) -> Iterator[Currency]:
        """
        Return a list of currencies for an alpha2 country code
        :params alpha2: alpha2 code of a country
        """
        try:
            country = Country(alpha2)
            return [Currency(cur) for cur in country.currencies()]
        except CurrencyNotFoundError as e:
            logging.error("Error fetching currency")
            logging.error(e)
            return []

    def get_rates(self, user: User = None, key: str = None, base_currency: str = None, start_date: date = None,
                  end_date: date = None) -> Iterator:
        """
        Return a list of rates for this currency and an optional base currency between two dates
        :params base_currency: code of the base currency
        :params start_date: rates from that date included
        :params end_date: rates to that date included
        :return: List of rates
        """
        from geocurrency.rates.models import Rate
        qs = Rate.objects.filter(currency=self.code)
        if user:
            qs = qs.filter(models.Q(user=user) | models.Q(user=None))
        if key:
            qs = qs.filter(key=key)
        if base_currency:
            qs = qs.filter(base_currency=base_currency)
        if start_date:
            qs = qs.filter(value_date__gte=start_date)
        if end_date:
            qs = qs.filter(value_date__lte=end_date)
        return qs

    @property
    def symbol(self):
        return CURRENCY_SYMBOLS.get(self.code, DEFAULT_SYMBOL)




from datetime import datetime, date, timedelta
from django.db import models
from django.http import HttpResponseForbidden
from django_filters import rest_framework as filters
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from rest_framework import viewsets, status, mixins
from rest_framework.decorators import action
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response
from rest_framework.status import HTTP_400_BAD_REQUEST
from rest_framework.views import APIView

from geocurrency.converters.serializers import ConverterResultSerializer
from geocurrency.currencies.models import Currency

from .filters import RateFilter
from .forms import RateForm
from .models import Rate, RateConverter
from .permissions import RateObjectPermission
from .serializers import RateSerializer, BulkSerializer, RateConversionPayloadSerializer


class RateViewSet(mixins.CreateModelMixin, mixins.ListModelMixin, mixins.RetrieveModelMixin, viewsets.GenericViewSet):
    queryset = Rate.objects.all()
    serializer_class = RateSerializer
    filter_backends = (filters.DjangoFilterBackend,)
    filterset_class = RateFilter
    pagination_class = PageNumberPagination
    permission_classes = [RateObjectPermission]
    display_page_controls = True

    def get_queryset(self):
        qs = super(RateViewSet, self).get_queryset()
        if self.request.user and self.request.user.is_authenticated:
            qs = qs.filter(
                models.Q(user=self.request.user) | models.Q(user__isnull=True)
            )
        else:
            qs = qs.filter(models.Q(user__isnull=True))
        return qs

    currency_latest_values = openapi.Parameter('currency_latest_values', openapi.IN_QUERY,
                                               description="Currency to filter on, limits results to latest values",
                                               type=openapi.TYPE_STRING)
    base_currency_latest_values = openapi.Parameter(
        'base_currency_latest_values', openapi.IN_QUERY,
        description="Base currency to filter on, limits results to latest values",
        type=openapi.TYPE_STRING)

    @swagger_auto_schema(manual_parameters=[currency_latest_values, base_currency_latest_values],
                         responses={200: RateSerializer})
    def list(self, request, *args, **kwargs):
        return super(RateViewSet, self).list(request, *args, **kwargs)

    def create(self, request, *args, **kwargs):
        rate_form = RateForm(request.data)
        if rate_form.is_valid():
            rate = rate_form.save(commit=False)
            if request.user and request.user.is_authenticated:
                rate.user = request.user
                rate.save()
                serializer = RateSerializer(rate)
                return Response(serializer.data, status=status.HTTP_201_CREATED)
            else:
                return HttpResponseForbidden()
        else:
            return Response(rate_form.errors, status=status.HTTP_400_BAD_REQUEST, content_type="application/json")

    @swagger_auto_schema(method='post', request_body=BulkSerializer, responses={201: RateSerializer})
    @action(['POST'], detail=False, url_path='bulk', url_name="bulk_create")
    def create_bulk(self, request):
        """
        Create rates for user on a range of dates
        """
        if not request.user or not request.user.is_authenticated:
            return HttpResponseForbidden()
        bs = BulkSerializer(data=request.data)
        if not bs.is_valid():
            return Response(bs.errors, status=status.HTTP_400_BAD_REQUEST)
        bulk_rate = bs.create(validated_data=bs.validated_data)
        rates = bulk_rate.to_rates(user=request.user)
        serializer = RateSerializer(rates, many=True)
        return Response(serializer.data, content_type="application/json", status=status.HTTP_201_CREATED)


class ConvertView(APIView):

    @swagger_auto_schema(request_body=RateConversionPayloadSerializer,
                         responses={200: ConverterResultSerializer})
    @action(['POST'], detail=False, url_path='', url_name="convert")
    def post(self, request, *args, **kwargs):
        """
        Converts a list of amounts with currency and date to a reference currency
        :param request: HTTP request
        """
        cps = RateConversionPayloadSerializer(data=request.data)
        if not cps.is_valid():
            return Response(cps.errors, status=HTTP_400_BAD_REQUEST, content_type="application/json")
        cp = cps.create(cps.validated_data)
        try:
            converter = RateConverter.load(cp.batch_id)
        except KeyError:
            converter = RateConverter(
                id=cp.batch_id,
                user=request.user,
                key=cp.key,
                base_currency=cp.target
            )
        if cp.data:
            if errors := converter.add_data(data=cp.data):
                return Response(errors, status=HTTP_400_BAD_REQUEST)
        if cp.eob or not cp.batch_id:
            result = converter.convert()
            serializer = ConverterResultSerializer(result)
            return Response(serializer.data, content_type="application/json")
        else:
            return Response({'id': converter.id, 'status': converter.status}, content_type="application/json")



import logging
from django.apps import AppConfig, apps
from django.conf import settings


class CountryConfig(AppConfig):
    name = 'countries'
    verbose_name = "Countries query app"

    def ready(self):
        super(CountryConfig, self).ready()






import uuid

import pint
from django.core.cache import cache
from django.test import TestCase
from django.utils.translation import ugettext_lazy as _
from rest_framework import status
from rest_framework.test import APIClient

from .models import UnitSystem, UnitConverter
from .serializers import UnitAmountSerializer


class UnitTest(TestCase):

    def test_creation(self):
        us = UnitSystem()
        unit = us.unit('meter')
        self.assertTrue(isinstance(unit.unit, pint.Unit))

    def test_readable_dimension(self):
        us = UnitSystem()
        unit = us.unit(unit_name='meter')
        self.assertEqual(unit.readable_dimension, _('length'))
        unit = us.unit(unit_name='US_international_ohm')
        self.assertEqual(unit.readable_dimension, f"{_('length')}^2 * {_('mass')} / {_('current')}^2 / {_('time')}^3")

    def test_list_request(self):
        client = APIClient()
        response = client.get(
            '/units/mks/units/'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_list_with_dimension_request(self):
        client = APIClient()
        response = client.get(
            '/units/mks/units/',
            data={
                'family': 'length'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(set([c['family'] for c in response.json()]), {'length'})

    def test_list_with_dimension_2_request(self):
        client = APIClient()
        response = client.get(
            '/units/mks/units/',
            data={
                'family': 'surface'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(set([c['family'] for c in response.json()]), {'surface'})

    def test_retrieve_request(self):
        client = APIClient()
        response = client.get(
            '/units/mks/units/meter/'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_compatible_request(self):
        client = APIClient()
        response = client.get(
            '/units/mks/units/meter/compatible/'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)


class UnitSystemTest(TestCase):

    def test_available_systems(self):
        us = UnitSystem()
        available_systems = us.available_systems()
        self.assertEqual(available_systems, ['Planck', 'SI', 'US', 'atomic', 'cgs', 'imperial', 'mks'])

    def test_available_units(self):
        us = UnitSystem(system_name='imperial')
        available_units = us.available_unit_names()
        self.assertIn('UK_hundredweight', available_units)

    def test_available_units_different(self):
        us = UnitSystem(system_name='mks')
        available_units = us.available_unit_names()
        self.assertIn('meter', available_units)
        self.assertNotIn('UK_hundredweight', available_units)
        imperial = UnitSystem(system_name='imperial')
        imperial_available_units = imperial.available_unit_names()
        self.assertIn('UK_hundredweight', imperial_available_units)

    def test_units_per_dimensionality(self):
        us = UnitSystem(system_name='mks')
        upd = us.units_per_dimensionality()
        self.assertIn(_('length'), upd)

    def test_dimensionalities(self):
        us = UnitSystem(system_name='mks')
        dims = us.dimensionalities
        self.assertIn(_('length'), dims)

    def test_list_request(self):
        client = APIClient()
        response = client.get(
            '/units/'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('SI', [us['system_name'] for us in response.json()])

    def test_retrieve_request(self):
        client = APIClient()
        response = client.get(
            '/units/SI/'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json().get("system_name"), "SI")
        self.assertIn("length", response.json().get("dimensions"))

    def test_retrieve_request_not_found(self):
        client = APIClient()
        response = client.get(
            '/units/sO/'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)


class UnitConverterTest(TestCase):
    base_system = 'SI'
    base_unit = 'meter'

    def setUp(self) -> None:
        self.converter = UnitConverter(base_system='SI', base_unit='meter')
        self.amounts = [
            {
                'system': 'SI',
                'unit': 'furlong',
                'value': 1,
                'date_obj': '2020-07-22'
            },
            {
                'system': 'SI',
                'unit': 'yard',
                'value': 1,
                'date_obj': '2020-07-22'
            },
        ]
        self.trash_amounts = [
            {
                'system': 'si',
                'unit': 'trop',
                'value': 100,
                'date_obj': '01/01/2020'
            },
            {
                'system': 'SI',
                'unit': 'trop',
                'value': 2,
                'date_obj': '2020-07-23'
            },
            {
                'system': 'SI',
                'unit': 'meter',
                'value': 'tete',
                'date_obj': '2020-07-23'
            },
            {
                'system': 'SI',
                'unit': 'meter',
                'value': 0,
                'date_obj': '01/01/2021'
            },
        ]

    def test_created(self):
        self.assertEqual(self.converter.status, self.converter.INITIATED_STATUS)

    def test_add_data(self):
        errors = self.converter.add_data(self.amounts)
        self.assertEqual(errors, [])
        self.assertEqual(self.converter.status, self.converter.INSERTING_STATUS)
        self.assertIsNotNone(cache.get(self.converter.id))

    def test_trash_amounts(self):
        converter = UnitConverter(base_system='SI', base_unit='meter')
        errors = converter.add_data(self.trash_amounts)
        self.assertEqual(len(errors), 3)
        self.assertIn("system", errors[0])
        self.assertIn("value", errors[1])
        self.assertIn("date_obj", errors[2])

    def test_add_empty_data(self):
        converter = UnitConverter(base_system='SI', base_unit='meter')
        errors = converter.add_data(data=None)
        self.assertEqual(len(errors), 1)

    def test_convert(self):
        result = self.converter.convert()
        self.assertEqual(result.id, self.converter.id)
        self.assertEqual(result.target, 'meter')
        self.assertEqual(self.converter.status, self.converter.FINISHED)
        self.assertEqual(len(result.errors), 0)
        self.assertEqual(len(result.detail), len(self.converter.data))
        converted_sum = sum([d.converted_value for d in result.detail])
        self.assertEqual(result.sum, converted_sum)

    def test_convert_request(self):
        amounts = UnitAmountSerializer(self.amounts, many=True)
        client = APIClient()
        response = client.post(
            '/units/convert/',
            data={
                'data': amounts.data,
                'base_system': 'SI',
                'base_unit': 'meter'
            },
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('sum', response.json())
        self.assertEqual(len(response.json().get('detail')), len(self.amounts))

    def test_convert_batch_request(self):
        batch_id = uuid.uuid4()
        client = APIClient()
        amounts = UnitAmountSerializer(self.amounts, many=True)
        response = client.post(
            '/units/convert/',
            data={
                'data': amounts.data,
                'base_system': 'SI',
                'base_unit': 'meter',
                'batch_id': batch_id,
            },
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('id', response.json())
        self.assertEqual(response.json().get('status'), UnitConverter.INSERTING_STATUS)
        self.assertEqual(response.json().get('id'), str(batch_id))
        response = client.post(
            '/units/convert/',
            data={
                'data': amounts.data,
                'batch_id': batch_id,
                'base_system': 'SI',
                'base_unit': 'meter',
                'eob': True
            },
            format='json')
        self.assertEqual(response.json().get('status'), UnitConverter.FINISHED)
        self.assertEqual(len(response.json().get('detail')), 2 * len(self.amounts))

    def test_watch_request(self):
        batch_id = uuid.uuid4()
        client = APIClient()
        amounts = UnitAmountSerializer(self.amounts, many=True)
        response = client.post(
            '/units/convert/',
            data={
                'data': amounts.data,
                'base_system': 'SI',
                'base_unit': 'meter',
                'batch_id': batch_id,
            },
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('id', response.json())
        self.assertEqual(response.json().get('status'), UnitConverter.INSERTING_STATUS)
        self.assertEqual(response.json().get('id'), str(batch_id))
        response = client.get(
            f'/watch/{str(batch_id)}/',
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json().get('status'), UnitConverter.INSERTING_STATUS)
        self.assertEqual(response.json().get('id'), str(batch_id))



from rest_framework import permissions


class CountryObjectPermission(permissions.BasePermission):

    def has_object_permission(self, request, view, obj):
        if request.method in permissions.SAFE_METHODS:
            return True
        return False



import logging
import requests
from django.conf import settings
from pycountry import countries

from . import Geocoder
from ..settings import *


class PeliasGeocoder(Geocoder):
    coder_type = 'pelias'
    server_url = None

    def __init__(self, server_url=None, *args, **kwargs):
        """
        Init pelias geocoder
        :params server_url: Custom pelias URL, defaults to 'http://127.0.0.1:3100/v1'
        """
        try:
            pelias_url = settings.PELIAS_GEOCODER_URL
        except AttributeError:
            pelias_url = GEOCODING_SERVICE_SETTINGS['pelias']['default_url']
        self.server_url = server_url or pelias_url

    def search(self, address, language=None, bounds=None, region=None, components=""):
        try:
            response = requests.get('{}/{}'.format(
                self.server_url,
                'search'

            ), {'text': address})
            data = response.json()
            if 'errors' in data:
                logging.error("Invalid request")
                logging.error(data.get('error'))
                return None
            return data
        except ValueError as e:
            logging.error("Invalid API configuration")
            logging.error(e)
        except IOError as e:
            logging.error("Invalid request")
            logging.error(e)
        return None

    def reverse(self, lat, lng):
        try:
            response = requests.get('{}/{}'.format(
                self.server_url,
                'reverse'
            ),
                {
                    'point.lat': lat,
                    'point.lon': lng,
                })
            data = response.json()
            if 'errors' in data:
                logging.error("ERROR - Invalid request")
                logging.error(data.get('error'))
                return None
            return data
        except ValueError:
            logging.error("Invalid API configuration")
        except IOError as e:
            logging.error("Invalid request", e)
        return None

    def parse_countries(self, data):
        """
        parse response from google service
        :params data: geocoding / reverse geocoding json
        :return: array of alpha2 codes
        """
        alphas = []
        if not data:
            return alphas
        for feature in data.get('features'):
            alphas.append(countries.get(alpha_3=feature.get('properties').get('country_a')).alpha_2)
        return alphas
